"""Custom logging formatters."""
# pylint: disable=no-self-use
import logging
import json
from contextlib import suppress
from time import gmtime
from time import strftime
from collections import OrderedDict
from pygments import highlight, lexers, formatters


# Use this when doing logs!!!
logger = logging.getLogger('t3')


class StandardFormatter(logging.Formatter):
    """Basic T3 Logging standard implementation."""

    def process(self, record):
        message = record.msg

        # Try to format message w/ args
        if record.args:
            with suppress(TypeError):
                message = message % record.args

        # Try to serialize message into a dict....
        # ...which will then be turned back into json
        with suppress(ValueError, TypeError):
            message = json.loads(message)

        out = OrderedDict([
            ('level', record.levelname),
            ('timestamp', self.format_time(record)),
            ('name', record.name),
            ('message', message),
            # ('path', OrderedDict([
            #     ('module', record.module),
            #     ('file_name', record.filename),
            #     ('line_number', record.lineno),
            #     ('function', record.funcName),
            #     ('path_name', record.pathname),
            # ])),
        ])

        # print(json_traceback())

        if hasattr(record, 'data'):
            out['data'] = record.data

        if record.exc_info or record.stack_info:
            info = record.exc_info if self.formatException(record.exc_info) else None
            stack = record.stack_info if self.formatStack(record.stack_info) else None
            out['exception'] = OrderedDict([
                ('info', info),
                ('stack', stack),
            ])

        return out

    def format(self, record):
        return self.process(record)

    def format_time(self, record, datefmt=None):
        """Format time with microseconds."""
        # pylint: disable=unused-argument
        intl_datetime = strftime('%Y-%m-%d %H:%M:%S', gmtime(record.created))
        return '{}.{}'.format(intl_datetime, int(record.msecs * 1000))


class JsonFormatter(StandardFormatter):
    """Uses python's json module to dump json string of record."""

    def format(self, record):
        try:
            return json.dumps(self.process(record), ensure_ascii=False)
        except Exception as e: # noqa
            return json.dumps({
                "level": 'ERROR',
                "message": 'The logger broke..... (╯°□°)╯︵ ┻━┻',
            }, ensure_ascii=False)


class JsonPrettyFormatter(StandardFormatter):
    """Uses python's pretty format module to pformat record."""

    def format(self, record):
        try:
            return highlight(
                json.dumps(self.process(record), indent=4, ensure_ascii=False),
                lexers.JsonLexer(),
                formatters.TerminalFormatter()
            )
        except Exception as e: # noqa
            return highlight(
                json.dumps({
                    "level": 'ERROR',
                    "message": 'The logger broke..... (╯°□°)╯︵ ┻━┻',
                }, ensure_ascii=False),
                lexers.JsonLexer(),
                formatters.TerminalFormatter()
            )



import logging
import pika

log = logging.getLogger(__name__)


class BasePublisher(object):
    """
    BasePublisher will run in a non-blocking way and close connection once the message has been sent.

    If RabbitMQ closes the connection, you should
    look at the output, as there are limited reasons why the connection may
    be closed, which usually are tied to permission related issues or
    socket timeouts.
    If the channel is closed, it will indicate a problem with one of the
    commands that were issued and that should surface in the output as well.
    """

    def __init__(self, amqp_url):
        """
        Create a new instance of the consumer class, passing in the AMQP URL used to connect to RabbitMQ.

        :param str amqp_url: The AMQP url to connect with
        """
        self._connection = None
        self._channel = None
        self._closing = False
        self._url = amqp_url

        self._exchange_name = ''
        self._exchange_type = 'direct'
        self._routing_key = ''
        self._message = ''

    def set_routing_key(self, routing_key):
        self._routing_key = routing_key

    def set_exchange(self, exchange_name, exchange_type='direct'):
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type

    def set_message(self, message):
        self._message = message

    def connect(self):
        """
        Connect to RabbitMQ, returning the connection handle.

        When the connection is established, the on_connection_open method
        will be invoked by pika.
        :rtype: pika.SelectConnection
        """
        log.info('Connecting to %s', self._url)
        return pika.SelectConnection(pika.URLParameters(self._url),
                                     self.on_connection_open,
                                     self.on_connection_error,   # !! Handle error opening connection
                                     stop_ioloop_on_close=False)

    def on_connection_open(self, unused_connection):
        """
        Call by pika once the connection to RabbitMQ has been established.

        It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :type unused_connection: pika.SelectConnection
        """
        log.info('Connection opened')
        self.add_on_connection_close_callback()
        self.open_channel()

    def on_connection_error(self, unused_connection, message):
        """
        Call by pika if the connection to RabbitMQ has failed to be established.

        It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :type unused_connection: pika.SelectConnection
        """
        log.error('Connection failed: %s', message)

    def add_on_connection_close_callback(self):
        """Add an on close callback when RabbitMQ closes the connection to the publisher unexpectedly."""
        log.info('Adding connection close callback')
        self._connection.add_on_close_callback(self.on_connection_closed)

    def on_connection_closed(self, connection, reply_code, reply_text):
        """
        Invoke by pika when the connection to RabbitMQ is closed unexpectedly.

        Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.
        :param pika.connection.Connection connection: The closed connection obj
        :param int reply_code: The server provided reply_code if given
        :param str reply_text: The server provided reply_text if given
        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            log.warning('Connection closed, reopening in 5 seconds: (%s) %s', reply_code, reply_text)

    def open_channel(self):
        """
        Open a new channel with RabbitMQ by issuing the Channel.Open RPC command.

        When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.
        """
        log.info('Creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """
        Invoke by pika when the channel has been opened.

        The channel object is passed in so we can make use of it.
        Since the channel is now open, we'll declare the exchange to use.
        :param pika.channel.Channel channel: The channel object
        """
        log.info('Channel opened')
        self._channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(self._exchange_name)

    def add_on_channel_close_callback(self):
        """Tell pika to call the on_channel_closed method if RabbitMQ unexpectedly closes the channel."""
        log.info('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reply_code, reply_text):
        """
        Invoke by pika when RabbitMQ unexpectedly closes the channel.

        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.
        :param pika.channel.Channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed
        """
        log.warning('Channel %i was closed: (%s) %s', channel, reply_code, reply_text)
        self._connection.close()

    def setup_exchange(self, exchange_name):
        """
        Start the exchange on RabbitMQ by invoking the Exchange.Declare RPC command.

        When it is complete, the on_exchange_declareok method will
        be invoked by pika.
        :param str|unicode exchange_name: The name of the exchange to declare
        """
        log.info('Declaring exchange %s', exchange_name)
        self._channel.exchange_declare(self.on_exchange_declareok,
                                       exchange_name,
                                       self._exchange_type)

    def on_exchange_declareok(self, unused_frame):
        """
        Invoke by pika when RabbitMQ has finished the Exchange.Declare RPC command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame
        """
        log.info('Exchange declared')
        self.start_publishing()

    def start_publishing(self):
        """Enable delivery confirmations and schedule the first message to be sent to RabbitMQ."""
        log.info(f'Sending message {self._message} to routing key {self._routing_key} '
                 f'on exchange {self._exchange_name}')
        self._connection.add_timeout(0.1, self._send_message)

    def _send_message(self):
        """Send message to the declared queue on the declared exchange."""
        self._channel.basic_publish(self._exchange_name, self._routing_key, self._message,
                                    pika.BasicProperties(content_type='text/plain', delivery_mode=1))
        log.info(f'Message {self._message} published, closing connection')
        self._connection.add_timeout(0.1, self.stop)

    def run(self):
        """Connect to RabbitMQ and then starting the IOLoop to block and allow the SelectConnection to operate."""
        self._connection = self.connect()
        self._connection.ioloop.start()

    def stop(self):
        """Close channels and connection, stop."""
        log.info('Sending a Basic.Cancel RPC command to RabbitMQ')
        self._closing = True
        log.info('Closing channel & connection')
        self._channel.close()
        self._connection.ioloop.stop()
        log.info('Stopped')



import logging
from collections import OrderedDict
from django.conf import settings
from rest_framework.views import exception_handler
from rest_framework.response import Response
from rest_framework import status

from t3.util import get_response_type
from t3.util import json_traceback


log = logging.getLogger(__name__)


def enveloped_exception_handler(exc, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)

    # Now add the HTTP status code to the response.
    if response is not None:
        try:
            response.data['status_code'] = response.status_code
        except:
            pass

    response = exception_handler(exc, context)
    if not response:
        response = Response({}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    response.data = OrderedDict([
        ('status', get_response_type(response.status_code)),
        ('status_code', response.status_code),
        ('message', str(exc)),
        ('data', json_traceback()),
    ])

    # Remove traceback if debug is off, or we don't have a 500 error
    if not settings.DEBUG or response.status_code != 500:
        del response.data['data']
    else:
        log.error(response.data['data'])

    return response



import requests
from django.conf import settings
import logging

from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

log = logging.getLogger(__name__)


def log_retry_obj(message, retries, factor, status_codes):
    log.info({
        'message': message,
        'retries': retries,
        'factor': factor,
        'statuses': ', '.join(status_codes),
    })


def get_retry_requests():
    # Check settings
    if not settings.EXTERNAL_RETRY:
        return None
    external_retry = settings.EXTERNAL_RETRY
    number = external_retry.get('number')
    factor = external_retry.get('factor')
    statuses = external_retry.get('status_codes')
    if (number or factor or statuses) is None:
        log_mess = 'Unable to use retry requests -> missing information'
        log_retry_obj(log_mess, number, factor, statuses.split(','))
        return None
    # Mount adapter
    session = requests.session()
    retry = Retry(
        total=int(number),
        read=int(number),
        connect=int(number),
        backoff_factor=float(factor),
        status_forcelist=tuple([int(s.strip()) for s in statuses.split(',')]),
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('https://', adapter)
    session.mount('http://', adapter)
    log_mess = 'Using retry requests'
    log_retry_obj(log_mess, number, factor, statuses.split(','))
    return session


# HTTP Get Request
def get(url, **kwargs):
    req = get_retry_requests()
    if req is not None:
        return req.get(url, **kwargs)
    return requests.get(url, **kwargs)


# HTTP Options Request
def options(url, **kwargs):
    req = get_retry_requests()
    if req is not None:
        return req.options(url, **kwargs)
    return requests.options(url, **kwargs)


# HTTP Head Request
def head(url, **kwargs):
    req = get_retry_requests()
    if req is not None:
        return req.head(url, **kwargs)
    return requests.head(url, **kwargs)


# HTTP Post Request
def post(url, **kwargs):
    req = get_retry_requests()
    if req is not None:
        return req.post(url, **kwargs)
    return requests.post(url, **kwargs)


# HTTP Put Request
def put(url, **kwargs):
    req = get_retry_requests()
    if req is not None:
        return req.put(url, **kwargs)
    return requests.put(url, **kwargs)


# HTTP Patch Request
def patch(url, **kwargs):
    req = get_retry_requests()
    if req is not None:
        return req.patch(url, **kwargs)
    return requests.patch(url, **kwargs)


# HTTP Delete Request
def delete(url, **kwargs):
    req = get_retry_requests()
    if req is not None:
        return req.delete(url, **kwargs)
    return requests.delete(url, **kwargs)



import os
import importlib
from distutils.util import strtobool
from t3.env import load_env
from t3.django.util import get_base_dir


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = get_base_dir()


# Get env dictionary.  This feeds from env, .env, and then vcap services
# For local dev, move .env.example to .env and play with settings
# In CI, a service.yaml is likely to be used
_env = load_env(dot_env_path=BASE_DIR)


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ['DJANGO_SECRET_KEY']

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = bool(strtobool(os.getenv('DJANGO_DEBUG', 'false')))

# Store this as a comma delimited list (eg. localhost,127.0.0.1)
ALLOWED_HOSTS = os.environ['DJANGO_ALLOWED_HOSTS'].split(',')


# Create of merge INSTALLED_APPS
# INSTALLED_APPS = globals().get('INSTALLED_APPS', [])
CORE_INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    # 3rd Party Packages
    'django_prometheus',
    'storages',
    'django_extensions',
    'corsheaders',
]


# Django Middleware
MIDDLEWARE = [
    'django_prometheus.middleware.PrometheusBeforeMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'corsheaders.middleware.CorsPostCsrfMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django_prometheus.middleware.PrometheusAfterMiddleware',

    # 3rd Party Middleware
    'log_request_id.middleware.RequestIDMiddleware',
]

# Django Templates
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


# Database
# https://docs.djangoproject.com/en/2.0/ref/settings/#databases
DATABASES = {
    'default': {
        'ENGINE': os.getenv('DB_ENGINE', 'django.db.backends.sqlite3'),
        'NAME': os.getenv('DB_NAME', os.path.join(BASE_DIR, 'db.sqlite3')),
        'USER': os.getenv('DB_USER'),
        'PASSWORD': os.getenv('DB_PASSWORD'),
        'HOST': os.getenv('DB_HOST'),
        'PORT': int(os.getenv('DB_PORT')) if os.getenv('DB_PORT') else None,
        'TEST': {
            'ENGINE': os.getenv('DB_TEST_ENGINE', 'django.db.backends.sqlite3'),
            'NAME': os.getenv('DB_TEST_NAME', os.path.join(BASE_DIR, 'db.sqlite3')),
            'USER': os.getenv('DB_TEST_USER'),
            'PASSWORD': os.getenv('DB_TEST_PASSWORD'),
            'HOST': os.getenv('DB_TEST_HOST'),
            'PORT': int(os.getenv('DB_TEST_PORT')) if os.getenv('DB_TEST_PORT') else None,
        }
    },
}

# Password validation
# https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

# Internationalization
# https://docs.djangoproject.com/en/2.0/topics/i18n/
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True


# ############################################################################
# T3 Specific Configs
# ############################################################################
try:
    PROJECT_PACKAGE = os.environ['DJANGO_SETTINGS_MODULE'].split('.')[0]
except KeyError:
    PROJECT_PACKAGE = None

LOGGING_LEVEL = os.getenv('DJANGO_LOGGING_LEVEL', 'INFO')

# Logging
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'request_id': {
            '()': 'log_request_id.filters.RequestIDFilter'
        }
    },
    'formatters': {
        'simple': {
            'format': u'%(asctime)s [%(levelname)-8s] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
        'django.server': {
            '()': 'django.utils.log.ServerFormatter',
            'format': u'%(asctime)s [%(levelname)-8s] %(message)s',
        },
        'json': {
            '()': 't3.django.log.DjangoJsonFormatter',
        },
        'json_pretty': {
            '()': 't3.django.log.DjangoJsonPrettyFormatter',
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'json_pretty',
            'stream': 'ext://sys.stdout',
            'filters': ['request_id'],
        },
    },
    'loggers': {
        't3': {
            'level': LOGGING_LEVEL,
            'handlers': ['console'],
            'propagate': False,
        },
        'django': {
            'level': LOGGING_LEVEL,
            'handlers': ['console'],
            'propagate': False,
        },
    },
}

if PROJECT_PACKAGE:
    LOGGING['loggers'][PROJECT_PACKAGE] = LOGGING['loggers']['t3']

# Log header stuff
# https://github.com/dabapps/django-log-request-id
LOG_REQUEST_ID_HEADER = "HTTP_X_REQUEST_ID"
GENERATE_REQUEST_ID_IF_NOT_IN_HEADER = True
REQUEST_ID_RESPONSE_HEADER = "RESPONSE_HEADER_NAME"
REQUEST_ID_RESPONSE_HEADER = "X-Response-Id"


# URL Prefix Stuff
URL_PREFIX = os.getenv('URL_PREFIX', '')
if URL_PREFIX != '':
    # Make sure URL_PREFIX ends with a '/'
    if not URL_PREFIX.endswith('/'):
        URL_PREFIX = URL_PREFIX + '/'
    # Make sure URL_PREFIX doesn't start with '/'
    if URL_PREFIX.startswith('/'):
        URL_PREFIX = URL_PREFIX[1:]


# ############################################################################
# AWS Configs
# ############################################################################
AWS_STORAGE = bool(strtobool(os.getenv('AWS_STORAGE', 'false')))
AWS_LOGGING = bool(strtobool(os.getenv('AWS_LOGGING', 'false')))

# Shared AWS Configs
if AWS_STORAGE or AWS_LOGGING:
    AWS_SERVICE_NAME = os.environ['AWS_SERVICE_NAME']
    AWS_ACCESS_KEY_ID = os.environ['AWS_ACCESS_KEY_ID']
    AWS_SECRET_ACCESS_KEY = os.environ['AWS_SECRET_ACCESS_KEY']
    AWS_REGION_NAME = os.environ['AWS_REGION_NAME']

# AWS S3 Storage
if AWS_STORAGE:
    AWS_STORAGE_BUCKET_NAME = os.environ['AWS_STORAGE_BUCKET_NAME']
    AWS_AUTO_CREATE_BUCKET = True
    AWS_S3_REGION_NAME = os.getenv('AWS_S3_REGION_NAME', AWS_REGION_NAME)

    AWS_S3_OBJECT_PARAMETERS = {
        'Expires': 'Thu, 31 Dec 2099 20:00:00 GMT',
        'CacheControl': 'max-age=94608000',
    }

    # Default media bucket
    DEFAULT_FILE_STORAGE = 't3.django.storages.MediaStorage'
    MEDIAFILES_BUCKET = os.getenv('AWS_MEDIAFILES_BUCKET', AWS_STORAGE_BUCKET_NAME)
    MEDIAFILES_LOCATION = os.getenv('AWS_MEDIAFILES_LOCATION', AWS_SERVICE_NAME + '/' + 'media')

    # Default static bucket
    STATICFILES_STORAGE = 't3.django.storages.StaticStorage'
    STATICFILES_BUCKET = os.getenv('AWS_STATICFILES_BUCKET', AWS_STORAGE_BUCKET_NAME)
    STATICFILES_LOCATION = os.getenv('AWS_STATICFILES_LOCATION', AWS_SERVICE_NAME + '/' + 'static')
else:
    # Static files (CSS, JavaScript, Images)
    # https://docs.djangoproject.com/en/2.0/howto/static-files/
    STATIC_URL = os.getenv('STATIC_URL', f'/{URL_PREFIX}static/')


# AWS CloudWatch Logging
if AWS_LOGGING:
    Session = importlib.import_module('boto3.session').Session

    # Add handler for CloudWatch (watchtower)
    AWS_DEFAULT_REGION = AWS_REGION_NAME
    LOGGING['handlers']['watchtower'] = {
        'formatter': 'json',
        'class': 'watchtower.django.DjangoCloudWatchLogHandler',
        'log_group': 'django',
        'stream_name': AWS_SERVICE_NAME,
        'filters': ['request_id'],
    }

    # Log using watchtower handler
    LOGGING['loggers']['t3']['handlers'] = ['console', 'watchtower']
    LOGGING['loggers']['django']['handlers'] = ['console', 'watchtower']

    if PROJECT_PACKAGE:
        LOGGING['loggers'][PROJECT_PACKAGE] = LOGGING['loggers']['t3']

# Django Extensions
SHELL_PLUS = "ipython"
IPYTHON_KERNEL_DISPLAY_NAME = "Django Shell-Plus"


# ############################################################################
# CORS Configs
# ############################################################################
# https://github.com/ottoyiu/django-cors-headers
# To see this in the header, then the request header must have `Origin` set
CORS = os.getenv('DJANGO_CORS', '*')
if CORS == '*':
    CORS_ORIGIN_ALLOW_ALL = True
else:
    CORS_ORIGIN_WHITELIST = ','.split(CORS)

# ############################################################################
# Proxy Configs
# ############################################################################
if bool(strtobool(os.getenv('DJANGO_SECURE_PROXY', 'false'))):
    USE_X_FORWARDED_HOST = True
    SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')


# ############################################################################
# Cache Configs (for calling external GET requests)
# Example:
#   REDIS_CACHE_REGEX: '((restaurants)|(products)).+((menu)|(modifiers))'
#   REDIS_CACHE_TTL: '4'
# Above parameters matches the url paths below with 4 minutes cache life:
#       * restaurants/23kasfe/menu
#       * products/234kasdf/modifiers
# Link: https://cachecontrol.readthedocs.io/en/latest/
# Link: https://regexr.com/ Regular Expression playground
# ############################################################################
REDIS_CACHE = bool(strtobool(os.getenv('REDIS_CACHE', 'false')))
if REDIS_CACHE:
    host = os.environ['REDIS_CACHE_HOST']
    port = os.getenv('REDIS_CACHE_PORT', '6379')
    database = os.getenv('REDIS_CACHE_DB', '0')
    password = os.getenv('REDIS_CACHE_PASS', '')
    passwordStr = f'?password={password}' if password else ''
    EXTERNAL_CACHE = {
        'redis': f'redis://{host}:{port}/{database}{passwordStr}',
        'pattern': os.environ['REDIS_CACHE_REGEX'],
        'minutes': os.environ['REDIS_CACHE_TTL'],
    }


# ############################################################################
# Retry Configs (for calling external REST API requests)
# Example:
#   EXTERNAL_RETRY_NUMBER: 3
#   EXTERNAL_RETRY_FACTOR: 0.3
#   EXTERNAL_RETRY_STATUS_CODES: 500, 502, 503, 504 - Comma separated list
# Above parameters retries the urls externally called for status codes
#   of 500, 502, 503, 504 which is basically server codes, if server is down
#   The above setup retries 3 times with a factor of 0.3.
#   First retry would be 0 seconds later,
#   Second retry would be 0.6 seconds later
#   Third retry would be 1.2 seconds later
# ############################################################################
EXTERNAL_RETRY = bool(strtobool(os.getenv('EXTERNAL_RETRY', 'false')))
if EXTERNAL_RETRY:
    EXTERNAL_RETRY = {
        'number': os.getenv('EXTERNAL_RETRY_NUMBER', '2'),
        'factor': os.getenv('EXTERNAL_RETRY_FACTOR', '0.1'),
        'status_codes': os.getenv('EXTERNAL_RETRY_STATUS_CODES', '500, 502'),
    }



# pylama:ignore=R0903,W0613,R0201
from rest_framework import status
from rest_framework.response import Response


class ListResourceMixin:
    def list(self, request, *args, **kwargs):
        resource = self.get_queryset()
        serializer = self.get_serializer(resource.list(), many=True)
        return Response(serializer.data)


class CreateResourceMixin:
    def create(self, request, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            # These kwargs will be passed in as extra validated data to create in the Resource
            serializer.save(**kwargs)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class RetrieveResourceMixin:
    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance.retrieve())
        return Response(serializer.data)


class UpdateResourceMixin:
    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        # Need to pass along the same keyword arguments for perform update
        # So that it will feed into the serializer, but more importantly the resource
        serializer.save(**kwargs)
        return Response(serializer.data)

    def partial_update(self, request, *args, **kwargs):
        kwargs['partial'] = True


class PartialUpdateResourceMixin:
    def partial_update(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=True)
        serializer.is_valid(raise_exception=True)
        # Need to pass along the same keyword arguments for perform update
        # So that it will feed into the serializer, but more importantly the resource
        serializer.save(**kwargs)
        return Response(serializer.data)


class DestroyResourceMixin:
    def destroy(self, request, **kwargs):
        resource = self.get_object()
        resource.destroy(**kwargs)
        return Response(status=status.HTTP_204_NO_CONTENT)



from abc import ABC
from collections import OrderedDict


class InstanceBase(ABC, OrderedDict):
    fields = []

    def __init__(self, *args, **kwds):
        # Fields defines the allowed data
        assert self.fields, \
            '`fields` is not set in {}'.format(self.__class__.__name__)

        try:
            initial_data = args[0]
        except IndexError:
            super().__init__(*args, **kwds)
            return

        # Verify we're only adding an allowed values
        for field_set in initial_data:
            assert field_set[0] in self.fields, \
                '`{}` is an invalid field.'.format(field_set)

        # Build OrderedDict structure based on `fields` & populate
        data = []
        for field in self.fields:
            try:
                value = list(filter(lambda x: x[0] == field, initial_data))[0][1]
            except IndexError:
                value = None

            data += [(field, value)]

        super().__init__(data, **kwds)

    def __setitem__(self, key, value, **kwargs):

        # Verify we're only adding an allowed value
        assert key in self.fields, '`{}` is an invalid key.'.format(key)

        super().__setitem__(key, value, **kwargs)

    @classmethod
    def factory(cls, *fields):
        # cls.fields = fields
        return type(
            cls.__name__,
            (cls, object),
            {'fields': fields}
        )
        # return cls


class ResourceBase(ABC):
    # response = None
    # raw = None
    data = {}
    fields = []

    instance_class = None
    # instance_class = InstanceBase.factory(fields)

    # def __init__(self):
    #     self.instance_class = InstanceBase.factory(fields)

    # def save_raw(self):
    #     self.raw = self.response.json()

    # def verify_status_code(self, code):
    #     if self.response.status_code != code:
    #         try:
    #             detail = 'Response json: {}'.format(self.response.json())
    #         except json.decoder.JSONDecodeError:
    #             detail = 'Response body: {}'.format(self.response.text)

    #         raise RequestException(
    #             detail=detail,
    #             status_code=self.response.status_code,
    #         )

    def get_instance(self, data):
        if not self.instance_class:
            self.instance_class = InstanceBase.factory(*self.fields)

        # assert self.instance_class.__name__ != 'InstanceBase', \
        #     '`instance_class` is InstanceBase, it must inherit from `InstanceBase`.'

        # assert issubclass(self.instance_class, InstanceBase), \
        #     '`instance_class` must inherit from `InstanceBase`.'

        return self.instance_class(data)

    # def process(self, code):
    #     self.verify_status_code(code)
    #     self.save_raw()



import sys
from os import environ
from os.path import dirname, abspath
from functools import lru_cache


@lru_cache()
def get_base_dir():
    """
    Use DJANGO_SETTINGS_MODULE to determine where the root project lives.

    This is used by Django's core code & is set in both manage.py & wsgi.py

    The assumption is that the path is `src.{project}.settings`
    """
    _settings_module = environ['DJANGO_SETTINGS_MODULE']
    _settings_file = sys.modules[_settings_module].__file__

    return dirname(dirname(dirname(abspath(_settings_file))))



import os
import logging
from t3.env import load_env

# Env vars
# Get env dictionary. This feeds from env, .env, and then vcap services
# For local dev, move .env.dist to .env and play with settings
# In CI, a service.yaml is likely to be used
BASE_DIR = os.getcwd()
_env = load_env(dot_env_path=BASE_DIR)


# ##########################################################################
# Event system configs & debug
#   DEBUG=true               - to set true or false to turn on/off debug logs
#   T3_EVENTS=false          - to set true or false to turn on/off events
#   T3_EVENTS_AMQP_URL       - rabbit mq amqp url string to connect to
# ##########################################################################

T3_EVENTS = True if str(os.getenv('T3_EVENTS', 'false')).lower() == 'true' else False
DEBUG = True if str(os.getenv('DEBUG', 'false')).lower() == 'true' else False

if DEBUG:
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.INFO)


def get_rabbit_mq_conn():
    if T3_EVENTS:
        return os.getenv('T3_EVENTS_AMQP_URL', 'amqp://rabbitmq:rabbitmq@localhost:5672/')
    else:
        raise Exception('T3 Events Disabled')


# Utility functions
def get_task_name(task_name):
    return f'task-{task_name}'


def get_topic_name(topic_name):
    return f'topic-{topic_name}'
